/* A ZPU emulator
 *
 * Based on an original idea by Toby Seckshund.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <assert.h>
#include <termio.h>

#include "zpu_memory.h"
#include "zpu.h"

uint32_t pc;
uint32_t sp;
uint32_t tos, nos;
uint8_t  instruction;
uint32_t cpu;

bool touchedPc = false;
bool decodeMask = false;

/*
 * BH: inline helper functions to make code easier to read
 *
 */

inline uint32_t pop() {
    sp+=4;
    return memoryReadLong(sp)
}

inline void push(uint32_t data) {
    memoryWriteLong(data,tos);
    sp-=4;
}

/*** end of helpers ***/

uint32_t flip(uint32_t i)
{
    uint32_t t = 0;
    for (int j = 0; j < 32; j++)
    {
        t |= ((i >> j) & 1) << (31 - j);
    }
    return t;
}


void emulate()
{
    sp = sp - 4; 
    memoryWriteLong(sp, pc + 1);
    pc = (memoryReadByte(pc) - 32) * VECTORSIZE + VECTORBASE;
    touchedPc = true;
}


#define SYS_READ  4
#define SYS_WRITE 5

void syscall()
{    
    int returnAdd = tos;
    int errNoAdd = memoryReadLong(sp + 4);
    int sysCallId = memoryReadLong(sp + 8);
    int fileNo = memoryReadLong(sp + 12);
    int charIndex = memoryReadLong(sp + 16);
    int stringLength = memoryReadLong(sp + 20);
    switch (sysCallId)
    {
	case SYS_WRITE:
	    for (int i = 0; i < stringLength; i++)
	    {
		putchar(memoryReadByte(charIndex++));
	    }	
	    // Return value via R0 (AKA memory address 0)
	    memoryWriteLong(0, stringLength);
	    break;
	case SYS_READ:
	    for (int i = 0; i < stringLength; i++)
	    {
		memoryWriteByte(charIndex++, tty_getchar());
	    }	
	    // Return value via R0 (AKA memory address 0)
	    memoryWriteLong(0, stringLength);
	    break;
    }    
}


static struct termio savemodes;
static int havemodes = 0;


int tty_break()
{
    struct termio modmodes;
    if(ioctl(fileno(stdin), TCGETA, &savemodes) < 0)
    {
      exit(1);
    }
    havemodes = 1;
    modmodes = savemodes;
    modmodes.c_lflag &= ~ICANON;
    modmodes.c_cc[VMIN] = 1;
    modmodes.c_cc[VTIME] = 0;
    return ioctl(fileno(stdin), TCSETAW, &modmodes);
}


int tty_getchar()
{
    return getchar();
}


int tty_fix()
{
    if(!havemodes)
    {
	return (0);
	return ioctl(fileno(stdin), TCSETAW, &savemodes);
    }
}


void zpu_reset()
{
    int i;
    if (tty_break() != 0)
    {
      return exit(1);
    }
//  tty_fix();
    memoryInitialize();
    pc = 0;
    touchedPc = true;
    sp = 0;
    tos = 0;
}


void zpu_execute()
{
    static int step = 0;
    for (;;)
    {
        touchedPc = false;

        instruction = memoryReadByte(pc);
//#define SINGLE_STEP
#ifdef SINGLE_STEP
	printf ("PC=%08x SP=%08x TOS=%08x OP=%02x DM=%02x debug=%08x\n", pc, sp, tos, instruction, decodeMask, 0);
	memoryDisplayLong(0 - 16*4, 16);
	printf("\n");
	getchar();
	if (step++ == 1000)
	{
	    printf("Done.\n");
	    fflush(0);	
	    exit(0);
	}
#endif
        if ((instruction & 0x80) == ZPU_IM)
        {
            if (decodeMask)
            {
                tos = tos << 7;
                tos = tos | instruction & 0x7f;
            }
            else
	    {   // push TOS, ignore wasted first four bytes for very first 
		push(tos);
		tos = instruction << 25;		// Sign extend
		tos = ((int32_t)tos) >> 25;
	    }
            decodeMask = true;
        }
        else
        {
            decodeMask = false;
            if ((instruction & 0xF0) == ZPU_ADDSP)
            {
                uint32_t addr;
                addr = instruction & 0x0F;
                addr = sp + addr * 4;
                tos = tos + memoryReadLong(addr);
            }
            else if ((instruction & 0xE0) == ZPU_LOADSP)
            {
                uint32_t addr;
                addr = (instruction & 0x1F) ^ 0x10; //BH: should XOR be sign extend?
                addr = sp + 4 * addr;
		push(tos);
                tos = memoryReadLong(addr);
            }
            else if ((instruction & 0xE0) == ZPU_STORESP)
            {
                uint32_t addr;
                addr = (instruction & 0x1F) ^ 0x10;
                addr = sp + 4 * addr;
                memoryWriteLong(addr, tos);
		tos = pop()
            }
            else
            {
                switch (instruction)
                {
                    case 0:
                        printf ("Breakpoint\n");
			printf ("PC=%08x SP=%08x  OP=%02x DM=%02x debug=%08x\n", pc, sp, instruction, decodeMask, 0);
			fflush(0);	
                        exit(1);
			break;
                    case ZPU_PUSHPC:
			push(tos);
			tos = pc;
                        break;
                    case ZPU_OR:
                    {
			nos = pop();
			tos = tos | nos;
                        break;
                    }
                    case ZPU_NOT:
                    {
			tos = ~tos;
                        break;
                    }
                    case ZPU_LOAD:
                    {
                        tos = memoryReadLong(tos);
                        break;
                    }
                    case ZPU_PUSHSPADD:
                    {
			tos =  (tos * 4) + sp;
                        break;
                    }
                    case ZPU_STORE:
		    {
			nos = pop()
                        memoryWriteLong(tos, nos);
			tos = pop();
                        break;
		    }
                    case ZPU_POPPC:
                    {
			pc = tos;
			tos = pop();
			touchedPc = true;
                        break;
                    }
                    case ZPU_POPPCREL:
                    {
                        pc = pc + tos;
			tos=pop();
                        touchedPc = true;
                        break;
                    }
                    case ZPU_FLIP:
                    {
			tos = flip(tos);
                        break;
                    }
                    case ZPU_ADD:
                    {
			nos = pop();
			tos = tos + nos;
                        break;
                    }
                    case ZPU_SUB:
                    {
			nos = pop();
			tos = nos - tos; //BH: as '-' is non-commutative note order of op
                        break;
                    }
                    case ZPU_PUSHSP:
                    {
			push(tos);
			tos = sp+4;
                        break;
                    }
                    case ZPU_POPSP:
                    {
			sp = tos;
			tos = pop();
                        break;
                    }
                    case ZPU_NOP:
                        break;
                    case ZPU_AND:
                    {
			nos = pop();
			tos &= nos;
                        break;
                    }
                    case ZPU_XOR:
                    {
			nos = pop();
			tos ^= nos;
                        break;
                    }
                    case ZPU_LOADB:
                    {
                        tos = memoryReadByte(tos);
                        break;
                    }
                    case ZPU_STOREB:
                    {
			nos = pop();
			memoryWriteByte(tos,nos);
			tos = pop();
                        break;
                    }
                    case ZPU_LOADH:
                    {
			printf("LOADH : Untested\n");
                        tos = memoryReadWord(tos);
                        break;
                    }
                    case ZPU_STOREH:
                    {
			printf("STOREH : Untested\n");
			nos = pop();
			memoryWriteWord(tos,nos);
			tos = pop();
                        break;
                    }
                    case ZPU_LESSTHAN:
                    {
			nos = pop();
			if ((int32_t)tos < (int32_t)nos)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_LESSTHANOREQUAL:
                    {
			nos = pop();
			if ((int32_t)tos <= (int32_t)nos)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_ULESSTHAN:
                    {
			nos = pop();
			if (tos < nos)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_ULESSTHANOREQUAL:
                    {
			nos = pop();
			if (tos <= nos)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_SWAP:
                    {
                        tos = ((tos >> 16) & 0xffff) | (tos << 16);
                        break;
                    }
                    case ZPU_MULT16X16:
                    {
			nos = pop();
                        tos = (nos & 0xffff) * (tos & 0xffff);
                        break;
                    }
                    case ZPU_EQBRANCH:
		    {
			nos = pop();
			if (nos == 0) {
                            pc = pc + tos;
                            touchedPc = true;
			}
			tos = pop();
                        break;
		    }
                    case ZPU_NEQBRANCH:
		    {
			nos = pop();
			if (nos != 0) {
                            pc = pc + tos;
                            touchedPc = true;
			}
			tos = pop();
                        break;
		    }
                    case ZPU_MULT:
		    {
			nos = pop();
			tos = (int32_t)tos * (int32_t)nos;
                        break;
		    }
                    case ZPU_DIV:
		    {
			nos = pop();
                        if (nos == 0)
                        {
                            printf("Divide by zero\n");
			    fflush(0);	
                            exit(1);
                        }
			tos = (int32_t)tos / (int32_t)nos;
                        break;
		    }
                    case ZPU_MOD:
		    {
			nos = pop();
                        if (nos == 0)
                        {
                            printf("Divide by zero\n");
			    fflush(0);	
                            exit(1);
                        }
			tos = (int32_t)tos % (int32_t)nos;
                        break;
		    }
                    case ZPU_LSHIFTRIGHT:
		    {
			nos = pop() & 0x3f;
			tos = nos >> tos;
                        break;
                    }
                    case ZPU_ASHIFTLEFT:
		    {
			nos = pop() & 0x3f;
			tos = nos << tos;
                        break;
                    }
                    case ZPU_ASHIFTRIGHT:
                    {
			nos = pop() & 0x3f;
			tos = (int32_t)nos >> tos;
                        break;
                    }
                    case ZPU_CALL:
		    {
			nos = tos;
			tos = pc + 1;
			pc  = nos;
                        touchedPc = true;
                        break;
                    }
                    case ZPU_CALLPCREL:
                    {
			nos = tos;
			tos = pc + 1;
			pc  = pc + nos;
                        touchedPc = true;
                        break;
                    }
                    case ZPU_EQ:
                    {
			nos = pop();
                        if (nos == tos)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_NEQ:
                    {
			nos = pop();
                        if (nos != tos)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_NEG:
                    {
			tos = -tos;
                        break;
                    }
                    case ZPU_CONFIG:
                    {
			cpu = tos;
			tos = pop();
                        printf ("CONFIG indicates CPU type is %d\n", cpu);
                        break;
		    }
                    case ZPU_SYSCALL:
			syscall();
                        break;
                    default:
                        printf ("Illegal Instruction\n");
			fflush(0);	
                        exit(1);
                        break;
                }
            }
        }
        if (!touchedPc)
        {
            pc = pc + 1;
            touchedPc = true;
        }
    }
}
